#include "SDL.h"

#include "sndserver.h"

#define DEFAULT_NUMCHANNELS 32


struct _SndChannel
{
	int len;
	int pos;
	int loop;
	int available;
	Uint8* buffer;
};

typedef struct _SndChannel SndChannel;

#define IsAvailable( channel) (channel->available != 0)
#define MustPlay( channel) (!IsAvailable(channel))
#define ReleaseChannel( channel) (channel->available = 1)
#define LockChannel( channel) ( channel->available = 0)

struct _SndData
{
	int len;
	Uint8* data;
	struct _SndData* next;
};

typedef struct _SndData SndData;

struct _SndServer
{
	SDL_AudioSpec spec;
	int numchannels;
	int numsnddata;
	SndChannel** channel;
	SndData* snddata;
};	
	
typedef struct _SndServer SndServer;


SndServer* sndserver = NULL;


static void FillBufferWithChannel( Uint8* buffer, 
                                   int buflen, 
								   SndChannel* channel)
{
	Uint8* channelbufpos = channel->buffer + channel->pos;
	int lentomix = channel->len - channel->pos;	



	if( lentomix <= buflen)
	{
		/* mixing everything */
		SDL_MixAudio( buffer, channelbufpos, lentomix, SDL_MIX_MAXVOLUME);

		/* mix a another chunck of channel if we must loop */
		if( channel->loop)
		{
			Uint8* bufpos = buffer + lentomix;
			lentomix = buflen - lentomix;
			SDL_MixAudio( bufpos, channel->buffer, lentomix, SDL_MIX_MAXVOLUME);

			/* set the new pos in channel */
			channel->pos = lentomix;

		}
		else
		{
			/* release the channel */
			ReleaseChannel( channel);
		}

	}
	else
	{
		/* mix to the end of buffer and refresh the channel pos */
		SDL_MixAudio( buffer, channelbufpos, buflen, SDL_MIX_MAXVOLUME);

		channel->pos += buflen;
	}

}


	


/* audio callback function */
static void FillBuffer( void * userdata, Uint8* buffer, int buflen)
{
	int i = 0;
	memset( buffer, 0, buflen);


	/* fill buffer with channels data */
	for( i = 0; i < sndserver->numchannels; ++i)
	{
		if( MustPlay( sndserver->channel[i]))
			FillBufferWithChannel( buffer, buflen, sndserver->channel[i]);
	}

}


SndChannel* NewSndChannel()
{
	SndChannel* ret = NULL;

	ret = (SndChannel*) malloc( sizeof( SndChannel));

	if( ret)
	{
		ret->len = 0;
		ret->pos = 0;
		ret->loop = 0;
		ret->available = 1;
		ret->buffer = NULL;
	}
	
	return ret;
}

void FreeChannel( SndChannel* sndchannel)
{
	free( sndchannel);
}


SndServer* CreateServer()
{
	SndServer* server = ( SndServer*) malloc( sizeof( SndServer));
	
	if( server)
	{
		int i = 0;
		
		/* this are default value for audio spec 
	 	* maybe later it's could become defined by user if needed*/ 
		server->spec.freq = 22050;
		server->spec.format = AUDIO_S16;
		server->spec.channels = 2;
		server->spec.samples = 512;
		server->spec.callback = FillBuffer;
		server->spec.userdata = NULL;

		/* create channels */
		server->numchannels = DEFAULT_NUMCHANNELS; /* hard defined but..*/ 
		server->channel = 
		(SndChannel**)malloc( server->numchannels * sizeof( SndChannel*));

		for( i = 0 ; i < server->numchannels; ++i)
			server->channel[i] = NewSndChannel();


		server->snddata = NULL;
		server->numsnddata = 0;

	}
	

	return server;
}	

int OpenSound()
{
	sndserver = CreateServer();
	
	if( !sndserver)
	{
		SDL_SetError( "Coudn't create sound server");

		return -1;
	}

	if( SDL_OpenAudio( &(sndserver->spec), NULL) < 0)
	{
		return -1;
	}

	return 0;
}

SndData* NewSndData()
{
	SndData* ret = NULL;

	ret = (SndData*) malloc( sizeof( SndData));

	ret->len = 0;
	ret->data = NULL;
	ret->next = NULL;

	return ret;
}

SndData* FreeSndData( SndData* snddata)
{
	SndData* ret = snddata->next;

	if( snddata->data)
	{
		SDL_FreeWAV( snddata->data);
	}

	free( snddata);

	return ret;
}

/* convert the SndData from the spec defined by sndspec into the server spec*/ 

void ConvertSndData( SndData* snddata, SDL_AudioSpec* sndspec)
{
	SDL_AudioCVT cvt;
	int newlen = 0;

	SDL_BuildAudioCVT( &cvt, 
	                   sndspec->format, 
	                   sndspec->channels,
					   sndspec->freq,
					   sndserver->spec.format,
					   sndserver->spec.channels,
					   sndserver->spec.freq);
	
	newlen =  snddata->len * cvt.len_mult;
	cvt.buf = (Uint8*) malloc( newlen * sizeof( Uint8));
	memcpy( cvt.buf, snddata->data, snddata->len);
	cvt.len = snddata->len;
	SDL_ConvertAudio( &cvt);
	SDL_FreeWAV( snddata->data);
	
	snddata->data = cvt.buf;
	snddata->len = newlen;	
}

int AddSndData( SndData* snddata)
{
	SndData* psnddata = sndserver->snddata;
	
	
	if( psnddata)
	{
		while( psnddata->next != NULL)
			psnddata = psnddata->next;
	
		psnddata->next = snddata;
	}
	else
		sndserver->snddata = snddata;

	
	return (sndserver->numsnddata++);
	
}

					

int AddSoundFile( const char* file)
{
	SDL_AudioSpec sndspec;
	
	SndData* snddata = NewSndData();
	
	if( SDL_LoadWAV( file, &sndspec, &(snddata->data), &snddata->len) == NULL)
	{
		FreeSndData( snddata);
		
		return (-1);
	}

	ConvertSndData( snddata, &sndspec);

	return AddSndData( snddata);
}


SndChannel* GetAvailableChannel()
{
	SndChannel* channel = NULL;
	int ch = 0;

	SDL_LockAudio();
	
	for( ch = 0; ch < sndserver->numchannels; ++ch)
	{
		if( IsAvailable( sndserver->channel[ch]))
			break;
	}

	if( ch < sndserver->numchannels)
	{
		channel = sndserver->channel[ch];
	}

	SDL_UnlockAudio();

	return channel;
}

SndData* GetSndDataFromID( int id)
{
	int i = 0;
	SndData* ret = sndserver->snddata;



	if( id < 0 || id >= sndserver->numsnddata)
		return NULL;
	
	for( i = 0; i < id ; ++i)
		ret = ret->next;
	
	return ret;
}



int PlaySound( int idsound, int loop)
{
	SndChannel* channel = GetAvailableChannel();
	SndData* snddata = GetSndDataFromID( idsound);

	if( !channel)
	{
		SDL_SetError( "All sound channels are occuped!");

		return (-1);
	}

	if( !snddata)
	{
		char errmsg[64];
		sprintf( errmsg, "Sound ID %d doesn't exist", idsound);
		SDL_SetError( errmsg);

		return (-1);
	}

	/* stop the callback during channel operation */
	SDL_LockAudio();

	/* init the channel with the sound data */
	channel->len = snddata->len;
	channel->buffer = snddata->data;
	
	/* set the playing variables */
	channel->pos = 0;
	channel->loop = loop;

	/* lock the channel */
	LockChannel( channel);

	/* restart the callback */
	SDL_UnlockAudio();

	return 0;
}

void CloseServer()
{
	int i = 0;
	SndData* psnddata = sndserver->snddata;

	for( i = 0; i < sndserver->numchannels; ++i)
		FreeChannel( sndserver->channel[i]);
	
	free( sndserver->channel);

	
	while( psnddata != NULL)
		psnddata = FreeSndData( psnddata);
		
}	
	

void CloseSound()
{
	if( sndserver)
		CloseServer();

	SDL_CloseAudio();
}


	
		

	




